@@ -11,8 +11,13 @@
 3. 为了使用方便，还提供了不需with的版本，可直接调用
 
 """
+from datetime import datetime
+from urllib.parse import urlparse
 
-from ..model import Format
+from loguru import logger
+
+from ..caller import call, check_methods
+from ..model import Format, RspModel
 from ..serializer.json_ import dict_to_json, json_to_dict, dict_to_obj
 from ..serializer.pkl_gzip import serialize, deserialize
 from ..utils.async_ import to_sync
@@ -51,23 +56,25 @@ def process_response(r):
 
 
 class WebSocketConnection:
-    def __init__(self, url, token=None):
+    def __init__(self, url: str, token=None):
+        path = urlparse(url).path
+        assert path.endswith(('/json', '/bytes', '/client', '/admin')), 'Python语言优先使用bytes，其它语言使用json'
+
+        if path.endswith('/json'):
+            self._fmt = Format.JSON
+        else:
+            self._fmt = Format.PKL_GZIP
+
         self._url = url
         self._token = token
         self._with = False  # 记录是否用了with
-        self._is_json = False  # 默认通讯格式设置，默认使用二进制通讯，减少类型等信息丢失
 
     async def __aenter__(self):
         """异步async with"""
         from websockets import connect
 
         headers = None if self._token is None else {"Authorization": f"Bearer {self._token}"}
-
-        if self._is_json:
-            self._client = connect(self._url + f'/json', extra_headers=headers)
-        else:
-            # 浏览器不支持headers，但写在query区又不安全 ?token={self._token}
-            self._client = connect(self._url + f'/bytes', extra_headers=headers)
+        self._client = connect(self._url, extra_headers=headers)
 
         self._ws = await self._client.__aenter__()
         self._with = True
@@ -96,16 +103,14 @@ class WebSocketConnection:
         if not self._with:
             await self.__aenter__()
 
-        if self._is_json:
-            fmt = Format.JSON
-        else:
-            fmt = Format.PKL_GZIP
+        # 如果是JSON格式可以考虑返回CSV
+        rsp_fmt = self._fmt
 
         d = dict(func=func, args=args, kwargs=kwargs,
-                 fmt=fmt,
+                 fmt=rsp_fmt.value,
                  cache_get=cache_get, cache_expire=cache_expire)
 
-        if fmt == Format.PKL_GZIP:
+        if self._fmt == Format.PKL_GZIP:
             # 二进制格式
             await self._ws.send(serialize(d).read())
         else:
@@ -114,3 +119,39 @@ class WebSocketConnection:
 
         rsp = await self._ws.recv()
         return process_response(rsp)
+
+    async def reverse(self):
+        """反弹RPC的被控端"""
+        while True:
+            req = await self._ws.recv()
+            req = deserialize(req)
+            logger.info(req)
+            # 不需要缓存
+            req['cache_get'] = False
+            req['cache_expire'] = 0
+            func = req['func']
+            try:
+                from ..config import METHODS_ALLOW, METHODS_BLOCK, METHODS_CHECK
+
+                # 反弹RPC权限要进行限制
+                if METHODS_CHECK:
+                    methods = func.split('.')
+                    if not check_methods(METHODS_ALLOW, methods, False):
+                        raise Exception(f'Method Not Allowed, {func} not in allowlist')
+                    if not check_methods(METHODS_BLOCK, methods, True):
+                        raise Exception(f'Method Not Allowed, {func} in blocklist')
+
+                key, buf, data = await call(**req)
+            except Exception as e:
+                # 主要是处理
+                key = type(e).__name__
+                # 这里没有缓存，因为这个错误是服务器内部检查
+                data = RspModel(status=401,  # status.HTTP_401_UNAUTHORIZED,
+                                datetime=datetime.now().isoformat(),
+                                func=func, args=req['args'], kwargs=req['args'])
+                data.type = type(e).__name__
+                data.data = repr(e)
+                data = data.dict()
+                buf = serialize(data).read()
+
+            await self._ws.send(buf)
